"""Configuration-related dataclasses."""

import logging
from dataclasses import dataclass, field
from typing import Callable, List, Optional

import backoff


@dataclass
class BackoffConfig:
    """Retry configuration."""

    wait_gen: Callable = field(default=backoff.expo)
    max_value: Optional[float] = None  # max duration for a single wait period
    max_tries: int = 10
    max_time: float = 60.0  # in seconds
    backoff_log_level: int = logging.DEBUG


@dataclass
class QueryConfig:
    """Query execution configiration."""

    # allows server-side parallel processing by issuing multiple reads concurrently for a single RPC
    num_sub_splits: int = 4

    # used to split the table into disjoint subsets of rows, to be processed concurrently using multiple RPCs
    # will be estimated from the table's row count, if not explicitly set
    num_splits: Optional[int] = None

    # each endpoint will be handled by a separate worker thread
    # a single endpoint can be specified more than once to benefit from multithreaded execution
    data_endpoints: Optional[List[str]] = None

    # a subsplit fiber will finish after sending this number of rows back to the client
    limit_rows_per_sub_split: int = 128 * 1024

    # each fiber will read the following number of rowgroups coninuously before skipping
    # in order to use semi-sorted projections this value must be 8 (this is the hard coded size of a row groups per row block).
    num_row_groups_per_sub_split: int = 8

    # can be disabled for benchmarking purposes
    use_semi_sorted_projections: bool = True

    # enforce using a specific semi-sorted projection (if enabled above)
    semi_sorted_projection_name: Optional[str] = None

    # used to estimate the number of splits, given the table rows' count
    rows_per_split: int = 4000000

    # used for worker threads' naming
    query_id: str = ""

    # non-negative integer, used for server-side prioritization of queued requests:
    # - requests with lower values will be served before requests with higher values.
    # - if unset, the request will be added to the queue's end.
    queue_priority: Optional[int] = None


@dataclass
class ImportConfig:
    """Import execution configiration."""

    import_concurrency: int = 2

    # import key column names
    key_names: Optional[List[str]] = None



# ruff: noqa: F841

import datetime as dt
import logging
from typing import Any, Dict, Optional, Sequence

import pyarrow as pa

import vastdb.bench.perf_bench.common.log_utils
from vastdb.bench.perf_bench.common import utils as bu
from vastdb.bench.perf_bench.common.constants import (
    DEFAULT_ARROW_KWARGS,
    VastConnDetails,
)
from vastdb.bench.perf_bench.query.arrow_common import build_arrow_filter

LOG = vastdb.bench.perf_bench.common.log_utils.get_logger(__name__)


# noinspection PyUnusedLocal
def query_pyarrow(
    universe: Sequence[str],
    columns: Optional[Sequence[str]] = None,
    from_t: Optional[dt.datetime] = None,
    to_t: Optional[dt.datetime] = None,
    path: Optional[str] = None,
    use_sid: bool = True,
    logger: Optional[logging.Logger] = None,
    backend_kwargs: Optional[Dict[str, Any]] = None,
) -> pa.Table:
    # ------------------------------------------------------------
    #    Query the PyArrow's parquet dataset
    # ------------------------------------------------------------
    backend_kwargs = backend_kwargs or {}
    conn_details: Optional[VastConnDetails] = backend_kwargs.pop("conn_details", None)
    conn_details = conn_details or VastConnDetails()
    fs_kwargs = {
        "botocore_client_kwargs": {
            "aws_access_key_id": conn_details.access,
            "aws_secret_access_key": conn_details.secret,
            "host": conn_details.s3_host,
        }
    }

    # Note that one can optimize the below by passing the filesystem object and/or partitioning
    #  scheme to avoid the slow penalty of discovering the partitioning upon init.
    dset = bu.get_parquet_dataset(path=path, filesystem=None, fs_kwargs=fs_kwargs)  # type: ignore[arg-type]

    ds_filter = build_arrow_filter(
        universe=universe,
        from_t=from_t,
        to_t=to_t,
        use_sid=use_sid,
    )

    # Cleanup the kwargs to be passed to the arrow scanner
    kwargs = backend_kwargs or DEFAULT_ARROW_KWARGS
    kwargs.pop("arrow_kwargs", None)
    if flt := kwargs.pop("filter", None):
        ds_filter = ds_filter & flt

    # Read the data
    # noinspection PyArgumentList
    table = dset.to_table(
        columns=columns,
        filter=ds_filter,
        **{**DEFAULT_ARROW_KWARGS, **(kwargs or {})},
    )

    return table



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class UInt16Literal(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = UInt16Literal()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsUInt16Literal(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # UInt16Literal
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # UInt16Literal
    def Value(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 0

def Start(builder): builder.StartObject(1)
def UInt16LiteralStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddValue(builder, value): builder.PrependUint16Slot(0, value, 0)
def UInt16LiteralAddValue(builder, value):
    """This method is deprecated. Please switch to AddValue."""
    return AddValue(builder, value)
def End(builder): return builder.EndObject()
def UInt16LiteralEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# Filter operation
class Filter(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Filter()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsFilter(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Filter
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # An identifiier for the relation. The identifier should be unique over the
    # entire plan. Optional.
    # Filter
    def Id(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vastdb.vast_flatbuf.org.apache.arrow.computeir.flatbuf.RelId import RelId
            obj = RelId()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Child relation
    # Filter
    def Rel(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vastdb.vast_flatbuf.org.apache.arrow.computeir.flatbuf.Relation import Relation
            obj = Relation()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # The expression which will be evaluated against input rows
    # to determine whether they should be excluded from the
    # filter relation's output.
    # Filter
    def Predicate(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vastdb.vast_flatbuf.org.apache.arrow.computeir.flatbuf.Expression import Expression
            obj = Expression()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def Start(builder): builder.StartObject(3)
def FilterStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddId(builder, id): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(id), 0)
def FilterAddId(builder, id):
    """This method is deprecated. Please switch to AddId."""
    return AddId(builder, id)
def AddRel(builder, rel): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(rel), 0)
def FilterAddRel(builder, rel):
    """This method is deprecated. Please switch to AddRel."""
    return AddRel(builder, rel)
def AddPredicate(builder, predicate): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(predicate), 0)
def FilterAddPredicate(builder, predicate):
    """This method is deprecated. Please switch to AddPredicate."""
    return AddPredicate(builder, predicate)
def End(builder): return builder.EndObject()
def FilterEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tabular

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class GetTableStatsResponse(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = GetTableStatsResponse()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsGetTableStatsResponse(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # GetTableStatsResponse
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # GetTableStatsResponse
    def NumRows(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # GetTableStatsResponse
    def SizeInBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # GetTableStatsResponse
    def IsExternalRowidAlloc(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # GetTableStatsResponse
    def AddressType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # GetTableStatsResponse
    def Vips(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from vastdb.vast_flatbuf.tabular.VipRange import VipRange
            obj = VipRange()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GetTableStatsResponse
    def VipsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # GetTableStatsResponse
    def VipsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # GetTableStatsResponse
    def SortingKeyEnabled(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # GetTableStatsResponse
    def SortingScore(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # GetTableStatsResponse
    def WriteAmplification(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # GetTableStatsResponse
    def AcummulativeRowInseritionCount(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

def GetTableStatsResponseStart(builder):
    builder.StartObject(9)

def Start(builder):
    GetTableStatsResponseStart(builder)

def GetTableStatsResponseAddNumRows(builder, numRows):
    builder.PrependInt64Slot(0, numRows, 0)

def AddNumRows(builder, numRows):
    GetTableStatsResponseAddNumRows(builder, numRows)

def GetTableStatsResponseAddSizeInBytes(builder, sizeInBytes):
    builder.PrependInt64Slot(1, sizeInBytes, 0)

def AddSizeInBytes(builder, sizeInBytes):
    GetTableStatsResponseAddSizeInBytes(builder, sizeInBytes)

def GetTableStatsResponseAddIsExternalRowidAlloc(builder, isExternalRowidAlloc):
    builder.PrependBoolSlot(2, isExternalRowidAlloc, 0)

def AddIsExternalRowidAlloc(builder, isExternalRowidAlloc):
    GetTableStatsResponseAddIsExternalRowidAlloc(builder, isExternalRowidAlloc)

def GetTableStatsResponseAddAddressType(builder, addressType):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(addressType), 0)

def AddAddressType(builder, addressType):
    GetTableStatsResponseAddAddressType(builder, addressType)

def GetTableStatsResponseAddVips(builder, vips):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(vips), 0)

def AddVips(builder, vips):
    GetTableStatsResponseAddVips(builder, vips)

def GetTableStatsResponseStartVipsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartVipsVector(builder, numElems):
    return GetTableStatsResponseStartVipsVector(builder, numElems)

def GetTableStatsResponseAddSortingKeyEnabled(builder, sortingKeyEnabled):
    builder.PrependBoolSlot(5, sortingKeyEnabled, 0)

def AddSortingKeyEnabled(builder, sortingKeyEnabled):
    GetTableStatsResponseAddSortingKeyEnabled(builder, sortingKeyEnabled)

def GetTableStatsResponseAddSortingScore(builder, sortingScore):
    builder.PrependInt64Slot(6, sortingScore, 0)

def AddSortingScore(builder, sortingScore):
    GetTableStatsResponseAddSortingScore(builder, sortingScore)

def GetTableStatsResponseAddWriteAmplification(builder, writeAmplification):
    builder.PrependInt64Slot(7, writeAmplification, 0)

def AddWriteAmplification(builder, writeAmplification):
    GetTableStatsResponseAddWriteAmplification(builder, writeAmplification)

def GetTableStatsResponseAddAcummulativeRowInseritionCount(builder, acummulativeRowInseritionCount):
    builder.PrependInt64Slot(8, acummulativeRowInseritionCount, 0)

def AddAcummulativeRowInseritionCount(builder, acummulativeRowInseritionCount):
    GetTableStatsResponseAddAcummulativeRowInseritionCount(builder, acummulativeRowInseritionCount)

def GetTableStatsResponseEnd(builder):
    return builder.EndObject()

def End(builder):
    return GetTableStatsResponseEnd(builder)



import concurrent.futures
import datetime as dt
import os
import time
from pathlib import Path
from typing import Optional, Sequence, Union

import numpy as np
import pandas as pd

from vastdb.bench.perf_bench.common.constants import (
    DEFAULT_END_T,
    DEFAULT_ROW_GROUP_SIZE,
    DEFAULT_START_T,
    DFAULT_PARQUET_COMPRESSION,
    LOCAL_FS_DS_PATH,
    ParquetCompression,
)
from vastdb.bench.perf_bench.common.log_utils import get_logger
from vastdb.bench.perf_bench.common.types import DateLikeT, PathLikeT
from vastdb.bench.perf_bench.common.utils import (
    get_parquet_dataset_root,
    get_session_minutes,
)
from vastdb.bench.perf_bench.dataset import secmaster as sm
from vastdb.bench.perf_bench.dataset.schemas import (
    BF,
    StockBarsPandasSchema,
)

LOG = get_logger(__name__)

DEFAULT_NUM_WORKERS = max(1, int((os.cpu_count() or 1) * 0.4))
DEFAULT_DS_ROOT_PATH = LOCAL_FS_DS_PATH

ref_bars: Optional[pd.DataFrame] = None


def generate_synthetic_stock_1m_bars_day(
    date: Optional[DateLikeT] = None,
    exchange: str = "XNYS",
) -> pd.DataFrame:
    sess_minutes = get_session_minutes(date=date or DEFAULT_START_T, exchange=exchange)
    uni_tickers = sm.large_uni
    sids = [sm.to_sid(t) for t in sm.large_uni]
    ref_px = pd.Series(
        [sm.get_indicative_px(sid) for sid in sids],
        dtype=BF.trade_close.pd_type,
    )
    uni_sz = len(uni_tickers)

    def _build_minute(m: pd.Timestamp) -> pd.DataFrame:
        return pd.DataFrame(
            {
                BF.sid.name: sids,
                BF.ts.name: [m.astimezone("UTC").tz_localize(None)] * uni_sz,  # type: ignore[arg-type]
                BF.ticker.name: uni_tickers,
                # Ask
                BF.ask_open.name: (
                    ao := ref_px * np.random.uniform(0.999, 0.001, uni_sz)
                ),
                BF.ask_high.name: ao * np.random.uniform(1.0, 1.001, uni_sz),
                BF.ask_low.name: ao * np.random.uniform(0.999, 1.0, uni_sz),
                BF.ask_close.name: ao * np.random.uniform(0.999, 1.001, uni_sz),
                BF.ask_qty.name: np.random.randint(1, 10000, uni_sz),
                # Bid
                BF.bid_open.name: (
                    bo := ref_px * np.random.uniform(0.999, 0.001, uni_sz)
                ),
                BF.bid_high.name: bo * np.random.uniform(1.0, 1.001, uni_sz),
                BF.bid_low.name: bo * np.random.uniform(0.999, 1.0, uni_sz),
                BF.bid_close.name: bo * np.random.uniform(0.999, 1.001, uni_sz),
                BF.bid_qty.name: np.random.randint(1, 10000, uni_sz),
                # Trades
                BF.trade_open.name: (
                    to := ref_px * np.random.uniform(0.999, 0.001, uni_sz)
                ),
                BF.trade_high.name: to * np.random.uniform(1.0, 1.001, uni_sz),
                BF.trade_low.name: to * np.random.uniform(0.999, 1.0, uni_sz),
                BF.trade_close.name: to * np.random.uniform(0.999, 1.001, uni_sz),
                BF.trade_volume.name: np.random.randint(1, 10000, uni_sz),
                # Vwap
                BF.vwap.name: ref_px * np.random.uniform(0.999, 1.001, uni_sz),
            }
        )

    # noinspection PyUnreachableCode
    return (
        pd.concat([_build_minute(m) for m in sess_minutes])
        .astype(StockBarsPandasSchema)
        .sort_values([BF.sid.name, BF.ts.name])
        .reset_index(drop=True)
    )


def worker_init():
    global ref_bars  # noqa: PLW0603
    ref_bars = generate_synthetic_stock_1m_bars_day()


# noinspection DuplicatedCode
def build_bars(
    dates: Union[Sequence[dt.date], pd.DatetimeIndex],
    output_dir: PathLikeT = DEFAULT_DS_ROOT_PATH,
    row_group_size: int = DEFAULT_ROW_GROUP_SIZE,
    compression: Union[str, ParquetCompression] = DFAULT_PARQUET_COMPRESSION,
):
    global ref_bars  # noqa: PLW0602
    if ref_bars is None:
        raise ValueError("Reference bars not initialized")

    output_dir = Path(str(output_dir))
    compression = ParquetCompression[compression]
    output_dir = get_parquet_dataset_root(output_dir, row_group_size, compression)
    output_dir.mkdir(parents=True, exist_ok=True)

    LOG.info(
        f"Building bars from {dates[0]:%Y-%m-%d} to {dates[-1]:%Y-%m-%d}, saving to {output_dir}"
    )

    # noinspection DuplicatedCode
    def build_new_bar(new_date: dt.date) -> pd.DataFrame:
        new_bars = ref_bars
        schema = new_bars.dtypes
        excluded_cols = {BF.ticker.value, BF.sid.value}
        for c in new_bars:
            if c in excluded_cols:
                continue
            kind: str = schema[c].kind  # type: ignore[call-overload]
            if kind == "f":
                new_bars[c] = new_bars[c] * np.random.uniform(0.8, 1.2)
            elif kind == "i":
                new_bars[c] = (new_bars[c] * np.random.uniform(0.2, 5)).astype(
                    schema[c]  # type: ignore[call-overload]
                )
            elif kind == "M":
                new_bars[c] = new_date + (new_bars[c] - new_bars[c].dt.normalize())  # type: ignore[operator]
        return new_bars

    for d in dates:
        try:
            start_t = time.time()

            bars_df = build_new_bar(new_date=d)
            generated_time = time.time() - start_t

            fpth = (
                output_dir
                / f"year={d:%Y}"
                / f"date={d:%Y%m%d}"
                / f"{d:%Y%m%d}_equity_etf_1m_bars.pq"
            )
            fpth.parent.mkdir(parents=True, exist_ok=True)

            # noinspection PyTypeChecker
            bars_df.to_parquet(
                path=fpth,
                engine="pyarrow",
                compression=compression,
                index=False,
                row_group_size=row_group_size,
            )  # type: ignore[call-overload]
            LOG.info(
                f"Written [{row_group_size=} {compression=}]"
                f"[total={(time.time() - start_t):.2f}, generate={generated_time:.2f}) seconds]: "
                f"{fpth}"
            )
        except Exception as e:
            LOG.exception(f"FAILURE with date {d}: {e}")


def common(output_dir: PathLikeT = DEFAULT_DS_ROOT_PATH):
    output_dir = Path(str(output_dir))
    output_dir.mkdir(parents=True, exist_ok=True)

    global ref_bars  # noqa: PLW0602
    if ref_bars is None:
        worker_init()


def generate_synthetic_stock_1m_bars(
    from_t: DateLikeT = DEFAULT_START_T,
    to_t: DateLikeT = DEFAULT_END_T,
    output_dir: PathLikeT = DEFAULT_DS_ROOT_PATH,
    row_group_size: int = DEFAULT_ROW_GROUP_SIZE,
    compression: Union[str, ParquetCompression] = DFAULT_PARQUET_COMPRESSION,
):
    common(output_dir=output_dir)
    dates = pd.bdate_range(
        pd.Timestamp(from_t).normalize(), pd.Timestamp(to_t).normalize()
    )
    build_bars(
        dates,
        output_dir=output_dir,
        row_group_size=row_group_size,
        compression=compression,
    )


def generate_concurrent_synthetic_stock_1m_bars(
    from_t: DateLikeT = DEFAULT_START_T,
    to_t: DateLikeT = DEFAULT_END_T,
    output_dir: PathLikeT = DEFAULT_DS_ROOT_PATH,
    num_workers: int = DEFAULT_NUM_WORKERS,
    row_group_size: int = DEFAULT_ROW_GROUP_SIZE,
    compression: Union[str, ParquetCompression] = DFAULT_PARQUET_COMPRESSION,
):
    if (num_workers := max(1, int(num_workers))) == 1:
        return generate_synthetic_stock_1m_bars(
            from_t=from_t,
            to_t=to_t,
            output_dir=output_dir,
            row_group_size=row_group_size,
            compression=compression,
        )

    common(output_dir=output_dir)
    dates = pd.bdate_range(
        pd.Timestamp(from_t).normalize(), pd.Timestamp(to_t).normalize()
    )
    batch_size = int(len(dates) / num_workers) + 1
    batches = [dates[i: i + batch_size] for i in range(0, len(dates), batch_size)]

    with concurrent.futures.ProcessPoolExecutor(
        max_workers=num_workers,
        initializer=worker_init,
    ) as executor:
        # Prepare a list of futures
        futures = [
            executor.submit(
                build_bars,
                dates=batch_dates,
                output_dir=output_dir,
                row_group_size=row_group_size,
                compression=compression,
            )
            for batch_dates in batches
        ]

        # Iterate over futures to get their results
        for future in concurrent.futures.as_completed(futures):
            LOG.info(future.result())



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# An expression representing a window function call.
class WindowCall(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = WindowCall()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsWindowCall(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # WindowCall
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # The expression to operate over
    # WindowCall
    def Expression(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vastdb.vast_flatbuf.org.apache.arrow.computeir.flatbuf.Expression import Expression
            obj = Expression()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # The kind of window frame
    # WindowCall
    def Kind(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Partition keys
    # WindowCall
    def Partitions(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from vastdb.vast_flatbuf.org.apache.arrow.computeir.flatbuf.Expression import Expression
            obj = Expression()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # WindowCall
    def PartitionsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # WindowCall
    def PartitionsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # Sort keys
    # WindowCall
    def Orderings(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from vastdb.vast_flatbuf.org.apache.arrow.computeir.flatbuf.SortKey import SortKey
            obj = SortKey()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # WindowCall
    def OrderingsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # WindowCall
    def OrderingsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # WindowCall
    def LowerBoundType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Lower window bound
    # WindowCall
    def LowerBound(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # WindowCall
    def UpperBoundType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Upper window bound
    # WindowCall
    def UpperBound(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

def Start(builder): builder.StartObject(8)
def WindowCallStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddExpression(builder, expression): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(expression), 0)
def WindowCallAddExpression(builder, expression):
    """This method is deprecated. Please switch to AddExpression."""
    return AddExpression(builder, expression)
def AddKind(builder, kind): builder.PrependUint8Slot(1, kind, 0)
def WindowCallAddKind(builder, kind):
    """This method is deprecated. Please switch to AddKind."""
    return AddKind(builder, kind)
def AddPartitions(builder, partitions): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(partitions), 0)
def WindowCallAddPartitions(builder, partitions):
    """This method is deprecated. Please switch to AddPartitions."""
    return AddPartitions(builder, partitions)
def StartPartitionsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def WindowCallStartPartitionsVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartPartitionsVector(builder, numElems)
def AddOrderings(builder, orderings): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(orderings), 0)
def WindowCallAddOrderings(builder, orderings):
    """This method is deprecated. Please switch to AddOrderings."""
    return AddOrderings(builder, orderings)
def StartOrderingsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def WindowCallStartOrderingsVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartOrderingsVector(builder, numElems)
def AddLowerBoundType(builder, lowerBoundType): builder.PrependUint8Slot(4, lowerBoundType, 0)
def WindowCallAddLowerBoundType(builder, lowerBoundType):
    """This method is deprecated. Please switch to AddLowerBoundType."""
    return AddLowerBoundType(builder, lowerBoundType)
def AddLowerBound(builder, lowerBound): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(lowerBound), 0)
def WindowCallAddLowerBound(builder, lowerBound):
    """This method is deprecated. Please switch to AddLowerBound."""
    return AddLowerBound(builder, lowerBound)
def AddUpperBoundType(builder, upperBoundType): builder.PrependUint8Slot(6, upperBoundType, 0)
def WindowCallAddUpperBoundType(builder, upperBoundType):
    """This method is deprecated. Please switch to AddUpperBoundType."""
    return AddUpperBoundType(builder, upperBoundType)
def AddUpperBound(builder, upperBound): builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(upperBound), 0)
def WindowCallAddUpperBound(builder, upperBound):
    """This method is deprecated. Please switch to AddUpperBound."""
    return AddUpperBound(builder, upperBound)
def End(builder): return builder.EndObject()
def WindowCallEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


#!/usr/bin/env python3

import functools
import itertools
import logging
import os
import random
import threading
import time
from concurrent.futures import ProcessPoolExecutor, as_completed

import numpy as np
import pyarrow as pa

import vastdb.errors
from vastdb.table import INTERNAL_ROW_ID
from vastdb.tests import metrics

logging.basicConfig(
    level="INFO",
    format="%(asctime)s %(levelname)-10s %(process)d/%(thread)d %(filename)s:%(lineno)d %(message)s")

log = logging.getLogger()

sdk_version = vastdb.version()
log.info("Python SDK version: %s", sdk_version)

NUM_COLUMNS = 10_000
COLUMNS_BATCH = 10

NUM_ROW_GROUPS = 100
ROW_GROUP_SIZE = 100_000


INTERNAL_ROWID_FIELD = pa.field(INTERNAL_ROW_ID, pa.uint64())  # used for UPDATE
EXTERNAL_ROWID_FIELD = pa.field("vastdb_rowid", pa.int64())  # used for INSERT & SELECT

SCHEMA = "perf"
TABLE = "sample"

SCHEMA_ARROW = pa.schema(
    [pa.field(f'c{i}', pa.float32()) for i in range(NUM_COLUMNS)]
)


def load_batch(bucket, session_kwargs, offset, limit):
    log.info('loading into [%d..%d)', offset, limit)

    # Iterate over all row-groups in this file
    rowids_range = range(offset, limit)
    rowids = pa.array(rowids_range, INTERNAL_ROWID_FIELD.type)

    session = vastdb.connect(**session_kwargs)
    metrics_rows = []

    with session.transaction() as tx:
        table = tx.bucket(bucket).schema(SCHEMA).table(TABLE)

        col = table[EXTERNAL_ROWID_FIELD.name]
        pred = (col >= rowids_range[0]) & (col <= rowids_range[-1])
        count = sum(len(rb) for rb in table.select(columns=[], predicate=pred))
        log.info("%d rows exist at %s", count, rowids_range)
        if count == len(rowids_range):
            # skip already loaded rows
            log.info('skipping [%d..%d)', offset, limit)

        pid = os.getpid()
        tid = threading.get_native_id()
        total_nbytes = 0
        calls = 0
        t0 = time.time()
        # Insert/update every chunk of columns in this rowgroup
        for j in range(0, len(SCHEMA_ARROW), COLUMNS_BATCH):
            cols_batch = list(SCHEMA_ARROW)[j:j + COLUMNS_BATCH]
            arrays = [
                pa.array(np.float32(np.random.uniform(size=[ROW_GROUP_SIZE])))
                for _ in cols_batch
            ]
            chunk = pa.table(data=arrays, schema=pa.schema(cols_batch))
            nbytes = chunk.get_total_buffer_size()
            start = time.perf_counter()
            if j == 0:
                chunk = chunk.add_column(0, EXTERNAL_ROWID_FIELD, rowids.cast(EXTERNAL_ROWID_FIELD.type))
                op = 'insert'
                table.insert(chunk)
            else:
                chunk = chunk.add_column(0, INTERNAL_ROWID_FIELD, rowids)
                op = 'update'
                table.update(chunk)
            finish = time.perf_counter()

            metrics_rows.append(metrics.Row(
                start=start, finish=finish, table_path=table.path, op=op,
                nbytes=nbytes, rows=len(chunk), cols=len(cols_batch),
                pid=pid, tid=tid, sdk_version=sdk_version))

            total_nbytes += nbytes
            calls += 1
            log.debug("%s into %s: %d rows x %d cols, %.3f MB",
                op, rowids_range, len(chunk), len(chunk.schema),
                chunk.get_total_buffer_size() / 1e6)

        dt = time.time() - t0

    log.info('loaded into [%d..%d): %d rows x %d cols, %.3f MB, %d RPCs, %.3f seconds',
             offset, limit, limit - offset, NUM_COLUMNS, total_nbytes / 1e6, calls, dt)
    return metrics_rows


def test_ingest(test_bucket_name, session_kwargs, tabular_endpoint_urls, num_workers, perf_metrics_db):
    session = vastdb.connect(**session_kwargs)
    metrics_table = metrics.Table(perf_metrics_db, "test_ingest")

    with session.transaction() as tx:
        b = tx.bucket(test_bucket_name)
        try:
            s = b.schema(SCHEMA)
        except vastdb.errors.MissingSchema:
            s = b.create_schema(SCHEMA)

        try:
            s.table(TABLE)
        except vastdb.errors.MissingTable:
            s.create_table(TABLE, pa.schema([EXTERNAL_ROWID_FIELD] + list(SCHEMA_ARROW)))

    ranges = [
        (i * ROW_GROUP_SIZE, (i + 1) * ROW_GROUP_SIZE)
        for i in range(NUM_ROW_GROUPS)
    ]

    with ProcessPoolExecutor(max_workers=num_workers) as executor:
        futures = [
            executor.submit(load_batch, test_bucket_name, session_kwargs | {'endpoint': url}, offset, limit)
            for (offset, limit), url in zip(ranges, itertools.cycle(tabular_endpoint_urls))
        ]
        log.info("spawned %d futures", len(futures))
        for future in as_completed(futures):
            metrics_table.insert(future.result())

    with session.transaction() as tx:
        t = tx.bucket(test_bucket_name).schema(SCHEMA).table(TABLE)
        count = sum(len(rb) for rb in t.select([]))
        log.info("%s has %d rows: %s", t, count, t.stats)


def run_query(session_kwargs, i, bucket_name, endpoint_url):
    num_columns = 2000
    row_groups_per_query = 10

    config = vastdb.table.QueryConfig(
        num_sub_splits=1,
        num_splits=1,
        limit_rows_per_sub_split=ROW_GROUP_SIZE,
        num_row_groups_per_sub_split=1)

    row_group_indices = list(range(NUM_ROW_GROUPS))
    r = random.Random(i)
    r.shuffle(row_group_indices)

    pid = os.getpid()
    tid = threading.get_native_id()
    metrics_rows = []

    session = vastdb.connect(**(session_kwargs | {"endpoint": endpoint_url}))
    with session.transaction() as tx:
        t = tx.bucket(bucket_name).schema(SCHEMA).table(TABLE)

        fields = list(t.arrow_schema)[1:]
        r.shuffle(fields)
        cols = [f.name for f in fields[:num_columns]]

        vastdb_rowid = t['vastdb_rowid']
        preds = []
        for offset in range(0, len(row_group_indices), row_groups_per_query):
            rowid_ranges = (
                vastdb_rowid.between(j * ROW_GROUP_SIZE, (j + 1) * ROW_GROUP_SIZE - 1)
                for j in row_group_indices[offset:offset + row_groups_per_query]
            )
            pred = functools.reduce((lambda x, y: x | y), rowid_ranges)
            preds.append(pred)

        for j, pred in enumerate(preds):
            log.info("%d) starting query #%d on %s", i, j, endpoint_url)

            start = time.perf_counter()
            res = t.select(columns=cols, predicate=pred, config=config)
            rows = 0
            data = 0
            for rb in res:
                rows += len(rb)
                data += rb.nbytes
                dt = time.perf_counter() - start
                log.info("%d) got query #%d batch %.3f[s], %.3f[GB] %.3f[MB/s], %.3f[Mrows]", i, j, dt, data / 1e9, data / 1e6 / dt, rows / 1e6)

            finish = time.perf_counter()
            dt = finish - start
            log.info("%d) finished query #%d %.3f[s], %.3f[GB], %.3f[MB/s], %.3f[Mrows]", i, j, dt, data / 1e9, data / 1e6 / dt, rows / 1e6)

            metrics_rows.append(metrics.Row(
                start=start, finish=finish, table_path=t.path, op="select",
                nbytes=data, rows=rows, cols=len(cols),
                pid=pid, tid=tid, sdk_version=sdk_version))

    return metrics_rows


def test_scan(test_bucket_name, session, num_workers, session_kwargs, tabular_endpoint_urls, perf_metrics_db):
    metrics_table = metrics.Table(perf_metrics_db, "test_scan")

    log.info("starting %d workers, endpoints=%s", num_workers, tabular_endpoint_urls)
    with ProcessPoolExecutor(max_workers=num_workers) as executor:
        futures = [
            executor.submit(run_query, session_kwargs, i, test_bucket_name, url)
            for i, url in zip(range(num_workers), itertools.cycle(tabular_endpoint_urls))
        ]
        for future in as_completed(futures):
            metrics_table.insert(future.result())

    log.info("finished %d workers", num_workers)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

# The varieties of relations
class RelationImpl(object):
    NONE = 0
    Aggregate = 1
    Filter = 2
    Join = 3
    Limit = 4
    LiteralRelation = 5
    OrderBy = 6
    Project = 7
    SetOperation = 8
    Source = 9




"""VAST Database bucket.

VAST S3 buckets can be used to create Database schemas and tables.
It is possible to list and access VAST snapshots generated over a bucket.
"""

import logging
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Iterable, Optional

from . import errors, schema, transaction

if TYPE_CHECKING:
    from .schema import Schema

log = logging.getLogger(__name__)


@dataclass
class Bucket:
    """VAST bucket."""

    name: str
    tx: "transaction.Transaction"
    _root_schema: "Schema" = field(init=False, compare=False, repr=False)

    def __post_init__(self):
        """Root schema is empty."""
        self._root_schema = schema.Schema(name="", bucket=self)

    def create_schema(self, name: str, fail_if_exists=True) -> "Schema":
        """Create a new schema (a container of tables) under this bucket."""
        return self._root_schema.create_schema(name=name, fail_if_exists=fail_if_exists)

    def schema(self, name: str, fail_if_missing=True) -> Optional["Schema"]:
        """Get a specific schema (a container of tables) under this bucket."""
        return self._root_schema.schema(name=name, fail_if_missing=fail_if_missing)

    def schemas(self, batch_size=None):
        """List bucket's schemas."""
        return self._root_schema.schemas(batch_size=batch_size)

    def snapshot(self, name: str, fail_if_missing=True) -> Optional["Bucket"]:
        """Get snapshot by name (if exists)."""
        snapshots, _is_truncated, _next_key = \
            self.tx._rpc.api.list_snapshots(bucket=self.name, name_prefix=name, max_keys=1)

        expected_name = f".snapshot/{name}"
        exists = snapshots and snapshots[0] == expected_name + "/"
        if not exists:
            if fail_if_missing:
                raise errors.MissingSnapshot(self.name, expected_name)
            else:
                return None

        return Bucket(name=f'{self.name}/{expected_name}', tx=self.tx)

    def snapshots(self) -> Iterable["Bucket"]:
        """List bucket's snapshots."""
        snapshots = []
        next_key = 0
        while True:
            curr_snapshots, is_truncated, next_key = \
                self.tx._rpc.api.list_snapshots(bucket=self.name, next_token=next_key)
            if not curr_snapshots:
                break
            snapshots.extend(curr_snapshots)
            if not is_truncated:
                break

        return [
            Bucket(name=f'{self.name}/{snapshot.strip("/")}', tx=self.tx)
            for snapshot in snapshots
        ]



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

# ----------------------------------------------------------------------
# Dictionary encoding metadata
# Maintained for forwards compatibility, in the future
# Dictionaries might be explicit maps between integers and values
# allowing for non-contiguous index values
class DictionaryKind(object):
    DenseArray = 0




from pathlib import Path
from typing import NamedTuple, TypedDict

import requests

from vastdb.bench.perf_bench.common.log_utils import get_logger

_MY_DIR = Path(__file__).parent
SM_PATH = _MY_DIR / "secmaster.py"

LOG = get_logger(__name__)


class NasdaqRawRecord(TypedDict):
    symbol: str
    name: str
    lastsale: str
    netchange: str
    pctchange: str
    marketCap: str
    url: str


class NasdaqRecord(NamedTuple):
    symbol: str
    id: int
    last_sale: float

    @staticmethod
    def id_from_ticker(ticker: str) -> int:
        base = 100
        max_width = 5
        offset = ord(" ")
        if len(ticker := ticker.strip()) > max_width:
            raise ValueError(f"Ticker too long: {ticker}")
        ticker = ticker.rjust(max_width, " ")
        return sum(
            (ord(c) - offset) * base ** (len(ticker) - i)
            for i, c in enumerate(ticker.upper())
        )

    @classmethod
    def from_raw_dict(cls, raw_dict: NasdaqRawRecord) -> "NasdaqRecord":
        return cls(
            symbol=(sym := raw_dict["symbol"].strip().upper()),
            id=cls.id_from_ticker(sym),
            last_sale=float(raw_dict["lastsale"].replace("$", "").replace(",", "")),
        )


def generate_secmaster():
    resp = requests.get(
        "http://api.nasdaq.com/api/screener/stocks?limit=3000",
        headers={"User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64)"},
    )
    resp.raise_for_status()
    resp_j = resp.json()
    records = [
        NasdaqRecord.from_raw_dict(r)
        for r in resp_j["data"]["table"]["rows"]
        if r and r["symbol"].strip()
    ]
    _ticker_to_sid = {r.symbol: r.id for r in records}
    _sid_to_ticker = {r.id: r.symbol for r in records}
    _indicative_px = {r.id: r.last_sale for r in records}

    uni = sorted(_ticker_to_sid)
    secm_file_contents = f"""
_ticker_to_sid = {_ticker_to_sid}


_sid_to_ticker = {_sid_to_ticker}


_indicative_px = {_indicative_px}


def to_sid(ticker: str) -> int:
    return _ticker_to_sid[ticker]


def to_ticker(sid: int) -> str:
    return _sid_to_ticker[sid]


def get_indicative_px(sid: int) -> float:
    return _indicative_px[sid]


UNI_SPEC = {{
    "Large": (large_uni := {uni}),
    "Single": large_uni[:1],
    "Tiny": (tiny_uni := large_uni[1::50]),
    "Small": (small_uni := large_uni[1::10]),
    "SmallSeq": large_uni[: len(small_uni)],
    "Medium": (med_uni := large_uni[1::6]),
    "MediumSeq": large_uni[: (len(med_uni))],
    "Medium2": (med2_uni := large_uni[1::8]),
    "Medium2Seq": large_uni[: (len(med2_uni))],
}}
"""
    with open(SM_PATH, "w") as f:
        f.write(secm_file_contents)

    LOG.info("Secmaster generated (total stocks: %d): %s", len(uni), SM_PATH)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

# Provided for forward compatibility in case we need to support different
# strategies for compressing the IPC message body (like whole-body
# compression rather than buffer-level) in the future
class BodyCompressionMethod(object):
    # Each constituent buffer is first compressed with the indicated
    # compressor, and then written with the uncompressed length in the first 8
    # bytes as a 64-bit little-endian signed integer followed by the compressed
    # buffer bytes (and then padding as required by the protocol). The
    # uncompressed length may be set to -1 to indicate that the data that
    # follows is not compressed, which can be useful for cases where
    # compression does not yield appreciable savings.
    BUFFER = 0




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tabular

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class GetRowColumnSecurityResponse(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = GetRowColumnSecurityResponse()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsGetRowColumnSecurityResponse(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # GetRowColumnSecurityResponse
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # GetRowColumnSecurityResponse
    def RowFilter(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from vastdb.vast_flatbuf.tabular.FilterString import FilterString
            obj = FilterString()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GetRowColumnSecurityResponse
    def RowFilterLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # GetRowColumnSecurityResponse
    def RowFilterIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # GetRowColumnSecurityResponse
    def ColumnMask(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from vastdb.vast_flatbuf.tabular.ColumnDetails import ColumnDetails
            obj = ColumnDetails()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GetRowColumnSecurityResponse
    def ColumnMaskLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # GetRowColumnSecurityResponse
    def ColumnMaskIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # GetRowColumnSecurityResponse
    def ColumnAllow(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from vastdb.vast_flatbuf.tabular.NameString import NameString
            obj = NameString()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GetRowColumnSecurityResponse
    def ColumnAllowLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # GetRowColumnSecurityResponse
    def ColumnAllowIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # GetRowColumnSecurityResponse
    def ColumnDeny(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from vastdb.vast_flatbuf.tabular.NameString import NameString
            obj = NameString()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GetRowColumnSecurityResponse
    def ColumnDenyLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # GetRowColumnSecurityResponse
    def ColumnDenyIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

def Start(builder): builder.StartObject(4)
def GetRowColumnSecurityResponseStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddRowFilter(builder, rowFilter): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(rowFilter), 0)
def GetRowColumnSecurityResponseAddRowFilter(builder, rowFilter):
    """This method is deprecated. Please switch to AddRowFilter."""
    return AddRowFilter(builder, rowFilter)
def StartRowFilterVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def GetRowColumnSecurityResponseStartRowFilterVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartRowFilterVector(builder, numElems)
def AddColumnMask(builder, columnMask): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(columnMask), 0)
def GetRowColumnSecurityResponseAddColumnMask(builder, columnMask):
    """This method is deprecated. Please switch to AddColumnMask."""
    return AddColumnMask(builder, columnMask)
def StartColumnMaskVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def GetRowColumnSecurityResponseStartColumnMaskVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartColumnMaskVector(builder, numElems)
def AddColumnAllow(builder, columnAllow): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(columnAllow), 0)
def GetRowColumnSecurityResponseAddColumnAllow(builder, columnAllow):
    """This method is deprecated. Please switch to AddColumnAllow."""
    return AddColumnAllow(builder, columnAllow)
def StartColumnAllowVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def GetRowColumnSecurityResponseStartColumnAllowVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartColumnAllowVector(builder, numElems)
def AddColumnDeny(builder, columnDeny): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(columnDeny), 0)
def GetRowColumnSecurityResponseAddColumnDeny(builder, columnDeny):
    """This method is deprecated. Please switch to AddColumnDeny."""
    return AddColumnDeny(builder, columnDeny)
def StartColumnDenyVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def GetRowColumnSecurityResponseStartColumnDenyVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartColumnDenyVector(builder, numElems)
def End(builder): return builder.EndObject()
def GetRowColumnSecurityResponseEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

# ----------------------------------------------------------------------
# Endianness of the platform producing the data
class Endianness(object):
    Little = 0
    Big = 1




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

# The kind of set operation being performed.
class SetOpKind(object):
    Union = 0
    Intersection = 1
    Difference = 2




# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# ----------------------------------------------------------------------
# A field represents a named column in a record / row batch or child of a
# nested type.
class Field(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Field()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsField(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Field
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Name is not required, in i.e. a List
    # Field
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Whether or not this field can contain nulls. Should be true in general.
    # Field
    def Nullable(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Field
    def TypeType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # This is the type of the decoded value if the field is dictionary encoded.
    # Field
    def Type(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # Present only if the field is dictionary encoded.
    # Field
    def Dictionary(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vastdb.vast_flatbuf.org.apache.arrow.flatbuf.DictionaryEncoding import DictionaryEncoding
            obj = DictionaryEncoding()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # children apply only to nested data types like Struct, List and Union. For
    # primitive types children will have length 0.
    # Field
    def Children(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from vastdb.vast_flatbuf.org.apache.arrow.flatbuf.Field import Field
            obj = Field()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Field
    def ChildrenLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Field
    def ChildrenIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # User-defined metadata
    # Field
    def CustomMetadata(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from vastdb.vast_flatbuf.org.apache.arrow.flatbuf.KeyValue import KeyValue
            obj = KeyValue()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Field
    def CustomMetadataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Field
    def CustomMetadataIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        return o == 0

def Start(builder): builder.StartObject(7)
def FieldStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddName(builder, name): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)
def FieldAddName(builder, name):
    """This method is deprecated. Please switch to AddName."""
    return AddName(builder, name)
def AddNullable(builder, nullable): builder.PrependBoolSlot(1, nullable, 0)
def FieldAddNullable(builder, nullable):
    """This method is deprecated. Please switch to AddNullable."""
    return AddNullable(builder, nullable)
def AddTypeType(builder, typeType): builder.PrependUint8Slot(2, typeType, 0)
def FieldAddTypeType(builder, typeType):
    """This method is deprecated. Please switch to AddTypeType."""
    return AddTypeType(builder, typeType)
def AddType(builder, type): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(type), 0)
def FieldAddType(builder, type):
    """This method is deprecated. Please switch to AddType."""
    return AddType(builder, type)
def AddDictionary(builder, dictionary): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(dictionary), 0)
def FieldAddDictionary(builder, dictionary):
    """This method is deprecated. Please switch to AddDictionary."""
    return AddDictionary(builder, dictionary)
def AddChildren(builder, children): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(children), 0)
def FieldAddChildren(builder, children):
    """This method is deprecated. Please switch to AddChildren."""
    return AddChildren(builder, children)
def StartChildrenVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def FieldStartChildrenVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartChildrenVector(builder, numElems)
def AddCustomMetadata(builder, customMetadata): builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(customMetadata), 0)
def FieldAddCustomMetadata(builder, customMetadata):
    """This method is deprecated. Please switch to AddCustomMetadata."""
    return AddCustomMetadata(builder, customMetadata)
def StartCustomMetadataVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def FieldStartCustomMetadataVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartCustomMetadataVector(builder, numElems)
def End(builder): return builder.EndObject()
def FieldEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tabular

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class VipRange(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = VipRange()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsVipRange(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # VipRange
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # VipRange
    def StartAddress(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # VipRange
    def AddressCount(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 0

def Start(builder): builder.StartObject(2)
def VipRangeStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddStartAddress(builder, startAddress): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(startAddress), 0)
def VipRangeAddStartAddress(builder, startAddress):
    """This method is deprecated. Please switch to AddStartAddress."""
    return AddStartAddress(builder, startAddress)
def AddAddressCount(builder, addressCount): builder.PrependUint16Slot(1, addressCount, 0)
def VipRangeAddAddressCount(builder, addressCount):
    """This method is deprecated. Please switch to AddAddressCount."""
    return AddAddressCount(builder, addressCount)
def End(builder): return builder.EndObject()
def VipRangeEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


"""VAST Database Python SDK."""

import functools

from . import session


# A helper function, useful as a short-hand for Session c-tor: `session = vastdb.connect(...)`
@functools.wraps(session.Session)
def connect(*args, **kwargs):  # noqa: D103
    return session.Session(*args, **kwargs)


def version():
    """Return VAST DB SDK version."""
    from importlib import metadata
    return metadata.distribution(__package__).version



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# Access a value for a given map key
class MapKey(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = MapKey()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsMapKey(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # MapKey
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Any expression can be a map key.
    # MapKey
    def Key(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vastdb.vast_flatbuf.org.apache.arrow.computeir.flatbuf.Expression import Expression
            obj = Expression()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def Start(builder): builder.StartObject(1)
def MapKeyStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddKey(builder, key): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(key), 0)
def MapKeyAddKey(builder, key):
    """This method is deprecated. Please switch to AddKey."""
    return AddKey(builder, key)
def End(builder): return builder.EndObject()
def MapKeyEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# Compressed Sparse Fiber (CSF) sparse tensor index.
class SparseTensorIndexCSF(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = SparseTensorIndexCSF()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsSparseTensorIndexCSF(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # SparseTensorIndexCSF
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # CSF is a generalization of compressed sparse row (CSR) index.
    # See [smith2017knl](http://shaden.io/pub-files/smith2017knl.pdf)
    #
    # CSF index recursively compresses each dimension of a tensor into a set
    # of prefix trees. Each path from a root to leaf forms one tensor
    # non-zero index. CSF is implemented with two arrays of buffers and one
    # arrays of integers.
    #
    # For example, let X be a 2x3x4x5 tensor and let it have the following
    # 8 non-zero values:
    # ```text
    #   X[0, 0, 0, 1] := 1
    #   X[0, 0, 0, 2] := 2
    #   X[0, 1, 0, 0] := 3
    #   X[0, 1, 0, 2] := 4
    #   X[0, 1, 1, 0] := 5
    #   X[1, 1, 1, 0] := 6
    #   X[1, 1, 1, 1] := 7
    #   X[1, 1, 1, 2] := 8
    # ```
    # As a prefix tree this would be represented as:
    # ```text
    #         0          1
    #        / \         |
    #       0   1        1
    #      /   / \       |
    #     0   0   1      1
    #    /|  /|   |    /| |
    #   1 2 0 2   0   0 1 2
    # ```
    # The type of values in indptrBuffers
    # SparseTensorIndexCSF
    def IndptrType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vastdb.vast_flatbuf.org.apache.arrow.flatbuf.Int import Int
            obj = Int()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # indptrBuffers stores the sparsity structure.
    # Each two consecutive dimensions in a tensor correspond to a buffer in
    # indptrBuffers. A pair of consecutive values at `indptrBuffers[dim][i]`
    # and `indptrBuffers[dim][i + 1]` signify a range of nodes in
    # `indicesBuffers[dim + 1]` who are children of `indicesBuffers[dim][i]` node.
    #
    # For example, the indptrBuffers for the above X is:
    # ```text
    #   indptrBuffer(X) = [
    #                       [0, 2, 3],
    #                       [0, 1, 3, 4],
    #                       [0, 2, 4, 5, 8]
    #                     ].
    # ```
    # SparseTensorIndexCSF
    def IndptrBuffers(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 16
            from vastdb.vast_flatbuf.org.apache.arrow.flatbuf.Buffer import Buffer
            obj = Buffer()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # SparseTensorIndexCSF
    def IndptrBuffersLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SparseTensorIndexCSF
    def IndptrBuffersIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # The type of values in indicesBuffers
    # SparseTensorIndexCSF
    def IndicesType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vastdb.vast_flatbuf.org.apache.arrow.flatbuf.Int import Int
            obj = Int()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # indicesBuffers stores values of nodes.
    # Each tensor dimension corresponds to a buffer in indicesBuffers.
    # For example, the indicesBuffers for the above X is:
    # ```text
    #   indicesBuffer(X) = [
    #                        [0, 1],
    #                        [0, 1, 1],
    #                        [0, 0, 1, 1],
    #                        [1, 2, 0, 2, 0, 0, 1, 2]
    #                      ].
    # ```
    # SparseTensorIndexCSF
    def IndicesBuffers(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 16
            from vastdb.vast_flatbuf.org.apache.arrow.flatbuf.Buffer import Buffer
            obj = Buffer()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # SparseTensorIndexCSF
    def IndicesBuffersLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SparseTensorIndexCSF
    def IndicesBuffersIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # axisOrder stores the sequence in which dimensions were traversed to
    # produce the prefix tree.
    # For example, the axisOrder for the above X is:
    # ```text
    #   axisOrder(X) = [0, 1, 2, 3].
    # ```
    # SparseTensorIndexCSF
    def AxisOrder(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # SparseTensorIndexCSF
    def AxisOrderAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # SparseTensorIndexCSF
    def AxisOrderLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SparseTensorIndexCSF
    def AxisOrderIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

def Start(builder): builder.StartObject(5)
def SparseTensorIndexCSFStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddIndptrType(builder, indptrType): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(indptrType), 0)
def SparseTensorIndexCSFAddIndptrType(builder, indptrType):
    """This method is deprecated. Please switch to AddIndptrType."""
    return AddIndptrType(builder, indptrType)
def AddIndptrBuffers(builder, indptrBuffers): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(indptrBuffers), 0)
def SparseTensorIndexCSFAddIndptrBuffers(builder, indptrBuffers):
    """This method is deprecated. Please switch to AddIndptrBuffers."""
    return AddIndptrBuffers(builder, indptrBuffers)
def StartIndptrBuffersVector(builder, numElems): return builder.StartVector(16, numElems, 8)
def SparseTensorIndexCSFStartIndptrBuffersVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartIndptrBuffersVector(builder, numElems)
def AddIndicesType(builder, indicesType): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(indicesType), 0)
def SparseTensorIndexCSFAddIndicesType(builder, indicesType):
    """This method is deprecated. Please switch to AddIndicesType."""
    return AddIndicesType(builder, indicesType)
def AddIndicesBuffers(builder, indicesBuffers): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(indicesBuffers), 0)
def SparseTensorIndexCSFAddIndicesBuffers(builder, indicesBuffers):
    """This method is deprecated. Please switch to AddIndicesBuffers."""
    return AddIndicesBuffers(builder, indicesBuffers)
def StartIndicesBuffersVector(builder, numElems): return builder.StartVector(16, numElems, 8)
def SparseTensorIndexCSFStartIndicesBuffersVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartIndicesBuffersVector(builder, numElems)
def AddAxisOrder(builder, axisOrder): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(axisOrder), 0)
def SparseTensorIndexCSFAddAxisOrder(builder, axisOrder):
    """This method is deprecated. Please switch to AddAxisOrder."""
    return AddAxisOrder(builder, axisOrder)
def StartAxisOrderVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def SparseTensorIndexCSFStartAxisOrderVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartAxisOrderVector(builder, numElems)
def End(builder): return builder.EndObject()
def SparseTensorIndexCSFEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# Order by relation
class OrderBy(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = OrderBy()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsOrderBy(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # OrderBy
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # An identifiier for the relation. The identifier should be unique over the
    # entire plan. Optional.
    # OrderBy
    def Id(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vastdb.vast_flatbuf.org.apache.arrow.computeir.flatbuf.RelId import RelId
            obj = RelId()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Child relation
    # OrderBy
    def Rel(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vastdb.vast_flatbuf.org.apache.arrow.computeir.flatbuf.Relation import Relation
            obj = Relation()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Define sort order for rows of output.
    # Keys with higher precedence are ordered ahead of other keys.
    # OrderBy
    def Keys(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from vastdb.vast_flatbuf.org.apache.arrow.computeir.flatbuf.SortKey import SortKey
            obj = SortKey()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # OrderBy
    def KeysLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # OrderBy
    def KeysIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

def Start(builder): builder.StartObject(3)
def OrderByStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddId(builder, id): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(id), 0)
def OrderByAddId(builder, id):
    """This method is deprecated. Please switch to AddId."""
    return AddId(builder, id)
def AddRel(builder, rel): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(rel), 0)
def OrderByAddRel(builder, rel):
    """This method is deprecated. Please switch to AddRel."""
    return AddRel(builder, rel)
def AddKeys(builder, keys): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(keys), 0)
def OrderByAddKeys(builder, keys):
    """This method is deprecated. Please switch to AddKeys."""
    return AddKeys(builder, keys)
def StartKeysVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def OrderByStartKeysVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartKeysVector(builder, numElems)
def End(builder): return builder.EndObject()
def OrderByEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# Optional compression for the memory buffers constituting IPC message
# bodies. Intended for use with RecordBatch but could be used for other
# message types
class BodyCompression(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = BodyCompression()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsBodyCompression(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # BodyCompression
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Compressor library.
    # For LZ4_FRAME, each compressed buffer must consist of a single frame.
    # BodyCompression
    def Codec(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # Indicates the way the record batch body was compressed
    # BodyCompression
    def Method(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

def Start(builder): builder.StartObject(2)
def BodyCompressionStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddCodec(builder, codec): builder.PrependInt8Slot(0, codec, 0)
def BodyCompressionAddCodec(builder, codec):
    """This method is deprecated. Please switch to AddCodec."""
    return AddCodec(builder, codec)
def AddMethod(builder, method): builder.PrependInt8Slot(1, method, 0)
def BodyCompressionAddMethod(builder, method):
    """This method is deprecated. Please switch to AddMethod."""
    return AddMethod(builder, method)
def End(builder): return builder.EndObject()
def BodyCompressionEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


import logging
import random
import time
from concurrent.futures import ThreadPoolExecutor

import pyarrow as pa

from vastdb.table import QueryConfig

logger = logging.getLogger(__name__)


def test_concurrent_query(session, test_bucket_name, schema_name, table_name):
    """
    This test runs several selective queries in parallel. It is used to check various internal VAST scenarios.
    """
    amount_of_queries_in_parallel = 10  # due to limit on requests connection-pool
    config = QueryConfig(num_splits=1, num_sub_splits=1)

    def _execute_single_query():
        with session.transaction() as tx:
            t = tx.bucket(test_bucket_name).schema(schema_name).table(table_name)
            pred = (t["a"] == 0)  # 0 is in the min-max range
            s = time.time()
            t.select(config=config, predicate=pred).read_all()
            e = time.time()
            logger.info(f"Query took {e - s}")

    logger.info(f"about to submit {amount_of_queries_in_parallel} queries in parallel")
    with ThreadPoolExecutor() as executor:
        futures = [executor.submit(_execute_single_query) for _ in range(amount_of_queries_in_parallel)]
        for future in futures:
            future.result()
    logger.info(f"finished running {amount_of_queries_in_parallel} queries")


def test_table_stats(session, test_bucket_name, schema_name, table_name):
    """
    Testing stats integrity while altering table
    """
    NUM_TIMES_TO_INSERT = 1000
    seed = random.randint(0, 10)
    logger.info(f"random seed is {seed}")
    r = random.Random(seed)

    with session.transaction() as tx:
        t = tx.bucket(test_bucket_name).schema(schema_name).table(table_name)
        initial_stat = t.get_stats()
        table_fields = t.columns()

    rand_values = {}  # create a dict with a random value from each column
    with session.transaction() as tx:
        t = tx.bucket(test_bucket_name).schema(schema_name).table(table_name)
        for col in table_fields:
            res = t.select(columns=[col.name]).read_all().column(col.name)
            rand_values[col.name] = res[int(r.uniform(0, len(res)))].as_py()

    logger.info(f"rand row to insert to the table - {rand_values}, {NUM_TIMES_TO_INSERT} times")
    rb = pa.RecordBatch.from_pylist([rand_values] * NUM_TIMES_TO_INSERT)
    with session.transaction() as tx:
        t = tx.bucket(test_bucket_name).schema(schema_name).table(table_name)
        t.insert(rb)
        time.sleep(2)  # waiting for stats to get updated
        new_stat = t.get_stats()

    logger.info("inserted to table")
    assert new_stat.size_in_bytes != initial_stat.size_in_bytes
    assert new_stat.num_rows - NUM_TIMES_TO_INSERT == initial_stat.num_rows


def test_ndu_while_querying(session, test_bucket_name, schema_name, table_name):
    """
    Executing queries while a NDU takes place.
    """
    amount_of_queries_in_parallel = 5
    config = QueryConfig(num_splits=1, num_sub_splits=1)
    logger.info(f'{test_bucket_name=}, {schema_name=}, {table_name=}')

    seed = random.randint(0, 100)
    logger.info(f'{seed=}')
    r = random.Random(seed)

    def _execute_single_query(expected=None):
        time.sleep(r.random())
        with session.transaction() as tx:
            t = tx.bucket(test_bucket_name).schema(schema_name).table(table_name)
            s = time.time()
            res = t.select(config=config).read_all()
            e = time.time()
            if expected:
                assert expected == res
            logger.info(f'query took {e - s}')
            return res

    initial_res = _execute_single_query()
    for i in range(300):
        with ThreadPoolExecutor() as executor:
            futures = [executor.submit(_execute_single_query, initial_res) for _ in range(amount_of_queries_in_parallel)]
            for future in futures:
                future.result()

        logger.info(f"finished running {amount_of_queries_in_parallel} queries, loop {i}")



import datetime as dt
from itertools import product
from typing import Dict, List, Optional, Sequence, Union

import pandas as pd

from vastdb.bench.perf_bench.common.constants import (
    DEFAULT_ARROW_KWARGS,
    DEFAULT_ROW_GROUP_SIZE,
    DEFAULT_START_T,
    DFAULT_PARQUET_COMPRESSION,
    ParquetCompression,
    VastConnDetails,
)
from vastdb.bench.perf_bench.common.types import PathLikeT
from vastdb.bench.perf_bench.common.utils import get_parquet_dataset_root, infer_fs_type
from vastdb.bench.perf_bench.dataset import secmaster as sm
from vastdb.bench.perf_bench.dataset.schemas import DEFAULT_BARS_COLUMNS
from vastdb.bench.perf_bench.orchestrate.scenario import BenchScenario
from vastdb.bench.perf_bench.query.query import QueryBackend


def generate_perf_bench_scenarios(
    base_key: str,
    conn_details: VastConnDetails,
    # Common parameters
    query_backends: Sequence[Union[str, QueryBackend]] = (QueryBackend.pyarrow,),
    universe_choices: Sequence[Union[str, Sequence[str]]] = ("Single",),
    columns_choices: Sequence[Sequence[str]] = (DEFAULT_BARS_COLUMNS,),
    dt_start_t: Union[dt.date, str] = DEFAULT_START_T,
    num_bdays: Sequence[int] = (1,),
    # Arrow-specific options
    fs_path_choices: Optional[Sequence[PathLikeT]] = None,
    rowgroup_size_choices: Optional[Sequence[int]] = None,
    compression_choices: Optional[Sequence[Union[str, ParquetCompression]]] = None,
    arrow_batching_spec_choices: Optional[Sequence[Dict[str, int]]] = None,
    # VastDB-specific options
    vdb_num_sub_splits_choices: Optional[Sequence[int]] = None,
    vdb_num_row_groups_per_sub_split_choices: Optional[Sequence[int]] = None,
) -> List[BenchScenario]:
    if not base_key:
        raise ValueError("base_key must be provided")

    uni_choices: List[List[str]] = []
    for k in universe_choices:
        if isinstance(k, str):
            uni_choices.append(sm.UNI_SPEC.get(k, [k]))
        else:
            uni_choices.append(list(k))

    columns_choices = [[c for c in cs if c] for cs in columns_choices if cs]
    dt_start_t = pd.Timestamp(dt_start_t)
    dt_range_choices = [
        (dt_start_t, dt_start_t + pd.tseries.offsets.BDay(d))
        for d in sorted(set(num_bdays))
        if d > 0
    ]

    scenarios = []
    for i, (  # type: ignore[misc]
        uni,
        columns,
        (from_t, to_t),
        path,
        rowgroup_size,
        compression,
        arrow_batch_spec,
        vdb_num_sub_splits,
        vdb_num_row_groups_per_sub_split,
        query_backend,
    ) in enumerate(
        product(
            uni_choices,
            columns_choices,
            dt_range_choices,
            # Arrow specific options
            (*(fs_path_choices or []), None),
            (*(rowgroup_size_choices or [DEFAULT_ROW_GROUP_SIZE]), None),
            (*(compression_choices or [DFAULT_PARQUET_COMPRESSION]), None),
            (*(arrow_batching_spec_choices or [DEFAULT_ARROW_KWARGS]), None),
            # VastDB specific options
            (*(vdb_num_sub_splits_choices or [1]), None),
            (*(vdb_num_row_groups_per_sub_split_choices or [8]), None),
            [QueryBackend[qb] for qb in query_backends],
        )
    ):
        if query_backend is QueryBackend.vastdb:
            if not all((vdb_num_sub_splits, vdb_num_row_groups_per_sub_split)) or any(
                (path, rowgroup_size, compression, arrow_batch_spec)
            ):
                # ignore path and rg size for vastdb runs
                continue
            backend_kwargs = {
                "num_sub_splits": vdb_num_sub_splits,
                "num_row_groups_per_sub_split": vdb_num_row_groups_per_sub_split,
                "conn_details": conn_details,
            }
            path = None
        elif query_backend is QueryBackend.pyarrow:
            if not all((path, rowgroup_size, compression, arrow_batch_spec)) or any(
                (vdb_num_sub_splits, vdb_num_row_groups_per_sub_split)
            ):
                # ignore non-vastdb runs without a path or rg size
                continue
            backend_kwargs: Dict[str, Union[str, int, VastConnDetails]] = arrow_batch_spec  # type: ignore[no-redef]
            backend_kwargs["conn_details"] = conn_details
            path = get_parquet_dataset_root(path, rowgroup_size, compression)  # type: ignore[arg-type]
        else:
            raise NotImplementedError(f"Unsupported query backend: {query_backend=!r}")

        scen = BenchScenario(
            key=f"{base_key}_{query_backend.name}",
            func=query_backend.run_query,
            kwargs=dict(
                universe=uni,
                columns=columns,
                from_t=from_t,
                to_t=to_t,
                path=path,
                backend_kwargs=backend_kwargs,
            ),
            default_result_params={
                "uni_sz": len(uni),  # type: ignore
                "num_days": (to_t - from_t).days,  # type: ignore
                "query_backend": query_backend.name,  # type: ignore
                "fs_type": infer_fs_type(path or ""),  # type: ignore
                "rowgroup_size": rowgroup_size or -1,  # type: ignore
                "arrow_fragment_readahead": (  # type: ignore
                    arrow_batch_spec["fragment_readahead"] if arrow_batch_spec else -1  # type: ignore
                ),
                "arrow_batch_readahead": (  # type: ignore
                    arrow_batch_spec["batch_readahead"] if arrow_batch_spec else -1  # type: ignore
                ),
                "arrow_batch_size": (  # type: ignore
                    arrow_batch_spec["batch_size"] if arrow_batch_spec else -1  # type: ignore
                ),
                "vdb_num_sub_splits": vdb_num_sub_splits or -1,  # type: ignore
                "vdb_num_row_groups_per_sub_split": vdb_num_row_groups_per_sub_split or -1,  # type: ignore
            },
        )

        scenarios.append(scen)

    return scenarios



import datetime as dt
import os
from typing import Dict, List, NamedTuple, Optional, Union

import pandas as pd

DEFAULT_CSV_SEPARATOR = ","


class BenchResult(NamedTuple):
    # The identifier for this result
    key: str
    round: int

    # Host details and time
    host: str
    pid: int

    # Results details
    start_ts: pd.Timestamp
    end_ts: pd.Timestamp
    duration_sec: float
    n_rows: int
    n_cols: int
    n_bytes: int

    params: Optional[Dict[str, Union[str, int, float]]] = None

    def to_dict(
        self,
        include_params: bool = True,
        flatten_params: bool = True,
    ) -> Dict[str, Union[str, int, float]]:
        dc = {f: getattr(self, f) for f in self._fields if f != "params"}
        if include_params:
            params = self.params or {}
            if flatten_params:
                dc = {**dc, **{pk: pv for pk, pv in params.items() if pk not in dc}}
            else:
                dc["params"] = params
        return dc

    def fields(
        self,
        include_params: bool = True,
        flatten_params: bool = True,
    ) -> List[str]:
        return list(
            self.to_dict(include_params=include_params, flatten_params=flatten_params)
        )

    def csv_header(
        self,
        separator: str = DEFAULT_CSV_SEPARATOR,
        include_params: bool = True,
    ) -> str:
        return separator.join(
            self.to_dict(include_params=include_params, flatten_params=True)
        )

    def to_csv(
        self,
        separator: str = DEFAULT_CSV_SEPARATOR,
        include_params: bool = True,
        include_header: bool = False,
    ) -> str:
        def _to_str(v) -> str:
            if isinstance(v, dt.date):
                v = v.isoformat()
            v_str = str(v) if not isinstance(v, str) else v
            if separator in v_str:
                if '"' in v_str:
                    raise ValueError(f"Can't handle double quotes in value: {v_str}")
                v_str = f'"{v_str}"'
            return v_str

        data = [
            _to_str(d)
            for d in self.to_dict(include_params=include_params, flatten_params=True)
        ]
        data_csv = separator.join(data)

        if include_header:
            data_csv = os.linesep.join(
                [
                    self.csv_header(separator=separator, include_params=include_params),
                    data_csv,
                ]
            )

        return data_csv



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tabular

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class GetProjectionTableStatsResponse(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = GetProjectionTableStatsResponse()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsGetProjectionTableStatsResponse(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # GetProjectionTableStatsResponse
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # GetProjectionTableStatsResponse
    def NumRows(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # GetProjectionTableStatsResponse
    def SizeInBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # GetProjectionTableStatsResponse
    def DirtyBlocksPercentage(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # GetProjectionTableStatsResponse
    def InitialSyncProgress(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

def Start(builder): builder.StartObject(4)
def GetProjectionTableStatsResponseStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddNumRows(builder, numRows): builder.PrependInt64Slot(0, numRows, 0)
def GetProjectionTableStatsResponseAddNumRows(builder, numRows):
    """This method is deprecated. Please switch to AddNumRows."""
    return AddNumRows(builder, numRows)
def AddSizeInBytes(builder, sizeInBytes): builder.PrependInt64Slot(1, sizeInBytes, 0)
def GetProjectionTableStatsResponseAddSizeInBytes(builder, sizeInBytes):
    """This method is deprecated. Please switch to AddSizeInBytes."""
    return AddSizeInBytes(builder, sizeInBytes)
def AddDirtyBlocksPercentage(builder, dirtyBlocksPercentage): builder.PrependInt64Slot(2, dirtyBlocksPercentage, 0)
def GetProjectionTableStatsResponseAddDirtyBlocksPercentage(builder, dirtyBlocksPercentage):
    """This method is deprecated. Please switch to AddDirtyBlocksPercentage."""
    return AddDirtyBlocksPercentage(builder, dirtyBlocksPercentage)
def AddInitialSyncProgress(builder, initialSyncProgress): builder.PrependInt64Slot(3, initialSyncProgress, 0)
def GetProjectionTableStatsResponseAddInitialSyncProgress(builder, initialSyncProgress):
    """This method is deprecated. Please switch to AddInitialSyncProgress."""
    return AddInitialSyncProgress(builder, initialSyncProgress)
def End(builder): return builder.EndObject()
def GetProjectionTableStatsResponseEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# Switch-style case expression
class SimpleCase(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = SimpleCase()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsSimpleCase(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # SimpleCase
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # The expression whose value will be matched
    # SimpleCase
    def Expression(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vastdb.vast_flatbuf.org.apache.arrow.computeir.flatbuf.Expression import Expression
            obj = Expression()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Matches for `expression`
    # SimpleCase
    def Matches(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from vastdb.vast_flatbuf.org.apache.arrow.computeir.flatbuf.CaseFragment import CaseFragment
            obj = CaseFragment()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # SimpleCase
    def MatchesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # SimpleCase
    def MatchesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # The default value if no cases match
    # SimpleCase
    def Else_(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vastdb.vast_flatbuf.org.apache.arrow.computeir.flatbuf.Expression import Expression
            obj = Expression()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def Start(builder): builder.StartObject(3)
def SimpleCaseStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddExpression(builder, expression): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(expression), 0)
def SimpleCaseAddExpression(builder, expression):
    """This method is deprecated. Please switch to AddExpression."""
    return AddExpression(builder, expression)
def AddMatches(builder, matches): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(matches), 0)
def SimpleCaseAddMatches(builder, matches):
    """This method is deprecated. Please switch to AddMatches."""
    return AddMatches(builder, matches)
def StartMatchesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def SimpleCaseStartMatchesVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartMatchesVector(builder, numElems)
def AddElse_(builder, else_): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(else_), 0)
def SimpleCaseAddElse_(builder, else_):
    """This method is deprecated. Please switch to AddElse_."""
    return AddElse_(builder, else_)
def End(builder): return builder.EndObject()
def SimpleCaseEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tabular

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class ImportDataRequest(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ImportDataRequest()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsImportDataRequest(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # ImportDataRequest
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ImportDataRequest
    def S3Files(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from vast_flatbuf.tabular.S3File import S3File
            obj = S3File()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ImportDataRequest
    def S3FilesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ImportDataRequest
    def S3FilesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # ImportDataRequest
    def KeyNames(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from vast_flatbuf.tabular.KeyName import KeyName
            obj = KeyName()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ImportDataRequest
    def KeyNamesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ImportDataRequest
    def KeyNamesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

def Start(builder): builder.StartObject(2)
def ImportDataRequestStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddS3Files(builder, s3Files): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(s3Files), 0)
def ImportDataRequestAddS3Files(builder, s3Files):
    """This method is deprecated. Please switch to AddS3Files."""
    return AddS3Files(builder, s3Files)
def StartS3FilesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ImportDataRequestStartS3FilesVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartS3FilesVector(builder, numElems)
def AddKeyNames(builder, keyNames): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(keyNames), 0)
def ImportDataRequestAddKeyNames(builder, keyNames):
    """This method is deprecated. Please switch to AddKeyNames."""
    return AddKeyNames(builder, keyNames)
def StartKeyNamesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ImportDataRequestStartKeyNamesVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartKeyNamesVector(builder, numElems)
def End(builder): return builder.EndObject()
def ImportDataRequestEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


import concurrent
import os
import socket
import time
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import pandas as pd

from vastdb.bench.perf_bench.orchestrate.bench_spec import BenchResult
from vastdb.bench.perf_bench.orchestrate.results_helpers import LOG


@dataclass
class BenchScenario:
    key: str
    func: Callable
    kwargs: Dict[str, Any]
    results: Optional[Tuple[BenchResult, ...]] = tuple()
    default_result_params: Optional[Dict[str, Union[str, int, float]]] = None

    # noinspection PyUnresolvedReferences
    def run(
        self,
        n_runs: int = 1,
        discard_first_run: bool = True,
        parallelism: int = 1,
        workers_init: Optional[Callable] = None,
    ):
        if parallelism == 1:
            self.results = tuple(
                self._do_run(n_runs=n_runs, discard_first_run=discard_first_run)
            )
        elif parallelism > 1:
            with concurrent.futures.ProcessPoolExecutor(
                max_workers=parallelism,
                initializer=workers_init,
            ) as executor:
                futures = [
                    executor.submit(
                        self._do_run, n_runs=n_runs, discard_first_run=discard_first_run
                    )
                    for _ in range(parallelism)
                ]

                # Iterate over futures to get their results
                results = []
                for future in concurrent.futures.as_completed(futures):
                    LOG.info("Future completed")
                    results.extend(future.result())
                self.results = tuple(results)
        else:
            raise ValueError("Can't use a negative number of parallel benches")

    def _do_run(
        self,
        n_runs: int = 1,
        discard_first_run: bool = True,
    ) -> List[BenchResult]:
        hostname = socket.gethostname()
        pid = os.getpid()
        results = []
        kwargs = self.kwargs or {}

        for i in range(n_runs + int(discard_first_run)):
            key = self.key

            start_ts = time.time_ns()
            ret_v = self.func(**kwargs)
            end_ts = time.time_ns()

            if discard_first_run and i == 0:
                LOG.info(f"Discarding first run ({key})")
                continue

            n_bytes, n_rows, n_cols = 0, 0, 0

            if hasattr(ret_v, "shape"):
                shape = ret_v.shape
                if len(shape):
                    n_rows = shape[0]
                if len(shape) > 1:
                    n_cols = shape[1]

            if hasattr(ret_v, "nbytes"):
                n_bytes = ret_v.nbytes
            elif hasattr(ret_v, "memory_usage"):
                n_bytes = ret_v.memory_usage().sum()
            elif hasattr(ret_v, "estimated_size"):
                n_bytes = ret_v.estimated_size()

            result = BenchResult(
                key=key,
                round=i,
                host=hostname,
                pid=pid,
                start_ts=pd.Timestamp(start_ts, unit="ns"),
                end_ts=pd.Timestamp(end_ts, unit="ns"),
                duration_sec=(end_ts - start_ts) / 1e9,
                n_rows=n_rows,
                n_cols=n_cols,
                n_bytes=n_bytes,
                params=self.default_result_params or {},
            )
            results.append(result)

            LOG.info(f"{result.duration_sec} sec")

        return results



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# ----------------------------------------------------------------------
# Data structures for dense tensors
# Shape data for a single axis in a tensor
class TensorDim(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = TensorDim()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsTensorDim(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # TensorDim
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Length of dimension
    # TensorDim
    def Size(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # Name of the dimension, optional
    # TensorDim
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def Start(builder): builder.StartObject(2)
def TensorDimStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddSize(builder, size): builder.PrependInt64Slot(0, size, 0)
def TensorDimAddSize(builder, size):
    """This method is deprecated. Please switch to AddSize."""
    return AddSize(builder, size)
def AddName(builder, name): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)
def TensorDimAddName(builder, name):
    """This method is deprecated. Please switch to AddName."""
    return AddName(builder, name)
def End(builder): return builder.EndObject()
def TensorDimEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


import contextlib
import logging
import threading
from http.server import BaseHTTPRequestHandler, HTTPServer
from itertools import cycle

import pytest

import vastdb.errors
from vastdb._internal import UnsupportedServer

log = logging.getLogger(__name__)


def test_hello_world(session):
    with session.transaction() as tx:
        assert tx.txid is not None


def test_bad_credentials(session):
    bad_session = vastdb.connect(access='BAD', secret='BAD', endpoint=session.api.url)
    with pytest.raises(vastdb.errors.Forbidden):
        with bad_session.transaction():
            pass


def test_bad_endpoint(session):
    backoff_config = vastdb.config.BackoffConfig(max_tries=3)
    with pytest.raises(vastdb.errors.ConnectionError):
        vastdb.connect(access='BAD', secret='BAD', endpoint='http://invalid-host-name-for-tests:12345', backoff_config=backoff_config)


def test_version_extraction():
    # A list of version and expected version parsed by API
    TEST_CASES = [
            ("nginx", UnsupportedServer),                               # non-vast server
            ("vast", NotImplementedError),                              # vast server without version in header
            ("vast 5", NotImplementedError),                            # major
            ("vast 5.2", NotImplementedError),                          # major.minor
            ("vast 5.2.0", NotImplementedError),                        # major.minor.patch
            ("vast 5.2.0.10", (5, 2, 0, 10)),                           # major.minor.patch.protocol
            ("vast 5.2.0.10 some other things", NotImplementedError),   # suffix
            ("vast 5.2.0.10.20", NotImplementedError),                  # extra version
    ]

    # Mock OPTIONS handle that cycles through the test cases response
    class MockOptionsHandler(BaseHTTPRequestHandler):
        versions_iterator = cycle(TEST_CASES)

        def __init__(self, *args) -> None:
            super().__init__(*args)

        def do_OPTIONS(self):
            self.send_response(204)
            self.end_headers()

        def version_string(self):
            return next(self.versions_iterator)[0]

        def log_message(self, format, *args):
            log.debug(format, *args)

    # start the server on localhost on some available port port
    server_address = ('localhost', 0)
    httpd = HTTPServer(server_address, MockOptionsHandler)

    def start_http_server_in_thread():
        log.info(f"Mock HTTP server is running on port {httpd.server_port}")
        httpd.serve_forever()
        log.info("Mock HTTP server killed")

    # start the server in a thread so we have the main thread to operate the API
    server_thread = threading.Thread(target=start_http_server_in_thread)
    server_thread.start()

    try:
        for _, expected in TEST_CASES:
            manager = contextlib.nullcontext()
            if isinstance(expected, type) and issubclass(expected, NotImplementedError):
                manager = pytest.raises(expected)
            with manager:
                s = vastdb.connect(endpoint=f"http://localhost:{httpd.server_port}", access="abc", secret="abc")
                assert s.api.vast_version == expected
    finally:
        # make sure we shut the server down no matter what
        httpd.shutdown()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tabular

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class FilterString(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = FilterString()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsFilterString(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # FilterString
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # FilterString
    def Filter(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def Start(builder): builder.StartObject(1)
def FilterStringStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddFilter(builder, filter): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(filter), 0)
def FilterStringAddFilter(builder, filter):
    """This method is deprecated. Please switch to AddFilter."""
    return AddFilter(builder, filter)
def End(builder): return builder.EndObject()
def FilterStringEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# Time is either a 32-bit or 64-bit signed integer type representing an
# elapsed time since midnight, stored in either of four units: seconds,
# milliseconds, microseconds or nanoseconds.
#
# The integer `bitWidth` depends on the `unit` and must be one of the following:
# * SECOND and MILLISECOND: 32 bits
# * MICROSECOND and NANOSECOND: 64 bits
#
# The allowed values are between 0 (inclusive) and 86400 (=24*60*60) seconds
# (exclusive), adjusted for the time unit (for example, up to 86400000
# exclusive for the MILLISECOND unit).
# This definition doesn't allow for leap seconds. Time values from
# measurements with leap seconds will need to be corrected when ingesting
# into Arrow (for example by replacing the value 86400 with 86399).
class Time(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Time()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsTime(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Time
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Time
    def Unit(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int16Flags, o + self._tab.Pos)
        return 1

    # Time
    def BitWidth(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 32

def Start(builder): builder.StartObject(2)
def TimeStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddUnit(builder, unit): builder.PrependInt16Slot(0, unit, 1)
def TimeAddUnit(builder, unit):
    """This method is deprecated. Please switch to AddUnit."""
    return AddUnit(builder, unit)
def AddBitWidth(builder, bitWidth): builder.PrependInt32Slot(1, bitWidth, 32)
def TimeAddBitWidth(builder, bitWidth):
    """This method is deprecated. Please switch to AddBitWidth."""
    return AddBitWidth(builder, bitWidth)
def End(builder): return builder.EndObject()
def TimeEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# An expression with an order
class SortKey(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = SortKey()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsSortKey(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # SortKey
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # SortKey
    def Expression(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vastdb.vast_flatbuf.org.apache.arrow.computeir.flatbuf.Expression import Expression
            obj = Expression()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # SortKey
    def Ordering(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

def Start(builder): builder.StartObject(2)
def SortKeyStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddExpression(builder, expression): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(expression), 0)
def SortKeyAddExpression(builder, expression):
    """This method is deprecated. Please switch to AddExpression."""
    return AddExpression(builder, expression)
def AddOrdering(builder, ordering): builder.PrependUint8Slot(1, ordering, 0)
def SortKeyAddOrdering(builder, ordering):
    """This method is deprecated. Please switch to AddOrdering."""
    return AddOrdering(builder, ordering)
def End(builder): return builder.EndObject()
def SortKeyEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


import datetime as dt
import logging
from typing import Any, Dict, Optional, Protocol, Sequence

import pyarrow as pa

from vastdb.bench.perf_bench.common.types import StrEnum


class QueryRunner(Protocol):
    def __call__(
        self,
        universe: Sequence[str],
        columns: Optional[Sequence[str]],
        from_t: Optional[dt.datetime],
        to_t: Optional[dt.datetime],
        path: Optional[str],
        use_sid: bool,
        logger: Optional[logging.Logger],
        backend_kwargs: Optional[Dict[str, Any]],
    ) -> pa.Table:
        ...


class QueryBackend(StrEnum):
    vastdb = "vastdb"
    pyarrow = "pyarrow"

    @property
    def _query_fun(self) -> QueryRunner:
        if self is QueryBackend.vastdb:
            from vastdb.bench.perf_bench.query.query_vastdb import query_vastdb

            return query_vastdb
        elif self is QueryBackend.pyarrow:
            from vastdb.bench.perf_bench.query.query_pyarrow import query_pyarrow

            return query_pyarrow
        raise NotImplementedError(f"Unsupported query backend: {self}")

    def run_query(self, *args, **kwargs) -> pa.Table:
        return self._query_fun(*args, **kwargs)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tabular

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class ListSchemasResponse(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ListSchemasResponse()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsListSchemasResponse(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # ListSchemasResponse
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ListSchemasResponse
    def BucketName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ListSchemasResponse
    def Schemas(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from vastdb.vast_flatbuf.tabular.ObjectDetails import ObjectDetails
            obj = ObjectDetails()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ListSchemasResponse
    def SchemasLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ListSchemasResponse
    def SchemasIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

def Start(builder): builder.StartObject(2)
def ListSchemasResponseStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddBucketName(builder, bucketName): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(bucketName), 0)
def ListSchemasResponseAddBucketName(builder, bucketName):
    """This method is deprecated. Please switch to AddBucketName."""
    return AddBucketName(builder, bucketName)
def AddSchemas(builder, schemas): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(schemas), 0)
def ListSchemasResponseAddSchemas(builder, schemas):
    """This method is deprecated. Please switch to AddSchemas."""
    return AddSchemas(builder, schemas)
def StartSchemasVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ListSchemasResponseStartSchemasVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartSchemasVector(builder, numElems)
def End(builder): return builder.EndObject()
def ListSchemasResponseEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Message(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Message()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsMessage(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Message
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Message
    def Version(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int16Flags, o + self._tab.Pos)
        return 0

    # Message
    def HeaderType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Message
    def Header(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # Message
    def BodyLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # Message
    def CustomMetadata(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from vastdb.vast_flatbuf.org.apache.arrow.flatbuf.KeyValue import KeyValue
            obj = KeyValue()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Message
    def CustomMetadataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Message
    def CustomMetadataIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

def Start(builder): builder.StartObject(5)
def MessageStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddVersion(builder, version): builder.PrependInt16Slot(0, version, 0)
def MessageAddVersion(builder, version):
    """This method is deprecated. Please switch to AddVersion."""
    return AddVersion(builder, version)
def AddHeaderType(builder, headerType): builder.PrependUint8Slot(1, headerType, 0)
def MessageAddHeaderType(builder, headerType):
    """This method is deprecated. Please switch to AddHeaderType."""
    return AddHeaderType(builder, headerType)
def AddHeader(builder, header): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(header), 0)
def MessageAddHeader(builder, header):
    """This method is deprecated. Please switch to AddHeader."""
    return AddHeader(builder, header)
def AddBodyLength(builder, bodyLength): builder.PrependInt64Slot(3, bodyLength, 0)
def MessageAddBodyLength(builder, bodyLength):
    """This method is deprecated. Please switch to AddBodyLength."""
    return AddBodyLength(builder, bodyLength)
def AddCustomMetadata(builder, customMetadata): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(customMetadata), 0)
def MessageAddCustomMetadata(builder, customMetadata):
    """This method is deprecated. Please switch to AddCustomMetadata."""
    return AddCustomMetadata(builder, customMetadata)
def StartCustomMetadataVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def MessageStartCustomMetadataVector(builder, numElems):
    """This method is deprecated. Please switch to Start."""
    return StartCustomMetadataVector(builder, numElems)
def End(builder): return builder.EndObject()
def MessageEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

# Various expression types
#
# WindowCall is a separate variant
# due to special options for each that don't apply to generic
# function calls. Again this is done to make it easier
# for consumers to deal with the structure of the operation
class ExpressionImpl(object):
    NONE = 0
    Literal = 1
    FieldRef = 2
    Call = 3
    ConditionalCase = 4
    SimpleCase = 5
    WindowCall = 6
    Cast = 7




"""Version-dependent features."""

import logging

from .errors import NotSupportedVersion

log = logging.getLogger()


class Features:
    """VAST database features - check if server is already support a feature."""

    def __init__(self, vast_version):
        """Save the server version."""
        self.vast_version = vast_version
        log.info("VAST version: %s", self.vast_version)

        self.check_imports_table = self._check(
            "Imported objects' table feature requires 5.2+ VAST release",
            vast_version >= (5, 2))

        self.check_return_row_ids = self._check(
            "Returning row IDs requires 5.1+ VAST release",
            vast_version >= (5, 1))

        self.check_enforce_semisorted_projection = self._check(
            "Semi-sorted projection enforcement requires 5.1+ VAST release",
            vast_version >= (5, 1))

        self.check_external_row_ids_allocation = self._check(
            "External row IDs allocation requires 5.1+ VAST release",
            vast_version >= (5, 1))

        self.check_elysium = self._check(
            "Elysium requires 5.3.5+ VAST release",
            vast_version >= (5, 3))  # TODO: make this validation stricter for v5.4 (beta/poc version is 5.3.0.x)

        self.check_zip_import = self._check(
            "Zip import requires 5.3.1+ VAST release",
            vast_version >= (5, 3, 1))

    def _check(self, msg, supported):
        log.debug("%s (current version is %s): supported=%s", msg, self.vast_version, supported)
        if not supported:
            def fail():
                raise NotSupportedVersion(msg, self.vast_version)
            return fail

        def noop():
            pass
        return noop



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuf

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# A cast expression
class Cast(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Cast()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsCast(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Cast
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # The expression to cast
    # Cast
    def Operand(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vastdb.vast_flatbuf.org.apache.arrow.computeir.flatbuf.Expression import Expression
            obj = Expression()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # The type to cast to. This value is a `Field` to allow complete representation
    # of arrow types.
    #
    # `Type` is unable to completely represent complex types like lists and
    # maps.
    # Cast
    def To(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from vastdb.vast_flatbuf.org.apache.arrow.flatbuf.Field import Field
            obj = Field()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def Start(builder): builder.StartObject(2)
def CastStart(builder):
    """This method is deprecated. Please switch to Start."""
    return Start(builder)
def AddOperand(builder, operand): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(operand), 0)
def CastAddOperand(builder, operand):
    """This method is deprecated. Please switch to AddOperand."""
    return AddOperand(builder, operand)
def AddTo(builder, to): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(to), 0)
def CastAddTo(builder, to):
    """This method is deprecated. Please switch to AddTo."""
    return AddTo(builder, to)
def End(builder): return builder.EndObject()
def CastEnd(builder):
    """This method is deprecated. Please switch to End."""
    return End(builder)


from pathlib import Path
from typing import List, Optional, Union

import numpy as np
import pandas as pd

from vastdb.bench.perf_bench.common.log_utils import get_logger
from vastdb.bench.perf_bench.orchestrate.bench_spec import BenchResult

LOG = get_logger(__name__)

PARAM_NAMES_SEP = "|"


def results_to_df(
    results: List[BenchResult],
) -> pd.DataFrame:
    results_df = pd.DataFrame([r.to_dict() for r in results])
    results_df["param_names"] = PARAM_NAMES_SEP.join(
        [c for c in (results[0].params or ()) if c in results_df]
    )
    return results_df


def save_results(
    results: List[BenchResult],
    results_path: Union[str, Path],
    append: bool = False,
):
    if not results:
        return

    results_path = Path(results_path)
    results_path.parent.mkdir(parents=True, exist_ok=True)
    results_df = results_to_df(results)

    if append and results_path.is_file():
        prev_results = pd.read_csv(results_path).astype(results_df.dtypes)
        results_df = pd.concat([prev_results, results_df], ignore_index=True)

    results_df.to_csv(results_path, index=False)
    LOG.info(f"Results written to {results_path}")


def calc_total_time_coverage_seconds(results_df: pd.DataFrame) -> float:
    # Step 1: Separate start and end times, tagging them differently
    starts = results_df["start_ts"].rename("time").to_frame().assign(event=1)
    ends = results_df["end_ts"].rename("time").to_frame().assign(event=-1)

    # Combine and sort chronologically
    timeline = pd.concat([starts, ends]).sort_values("time")

    # Step 2 & 3: Calculate cumulative sum to identify active intervals
    timeline["active_intervals"] = timeline["event"].cumsum()

    # Identify transitions (changes in active_intervals)
    timeline["change"] = timeline["active_intervals"].diff().fillna(1).astype(bool)

    # Calculate durations between changes
    timeline["duration"] = timeline["time"].diff().shift(-1).fillna(pd.Timedelta(0))

    # Only consider periods where at least one interval is active
    total_coverage: pd.Timedelta = timeline.query("active_intervals > 0")[
        "duration"
    ].sum()
    return total_coverage.total_seconds()


def calculate_aggregate_stats(
    results: Optional[List[BenchResult]] = None,
    results_df: Optional[pd.DataFrame] = None,
    results_path: Union[None, str, Path] = None,
) -> pd.DataFrame:
    if sum(bool(x) for x in (results, results_df, results_path)) != 1:
        raise ValueError(
            "Exactly one of results, results_df, or results_path must be provided"
        )

    if results:
        r_df = results_to_df(results)
        group_flds = list(results[0].params or ())
    else:
        r_df: pd.DataFrame = (  # type: ignore[assignment,no-redef]
            pd.read_csv(results_path) if results_path else results_df
        )
        p_names: pd.Series = results_df["param_names"]  # type: ignore[index]
        group_flds = list(p_names.values[0].split(PARAM_NAMES_SEP))
    group_flds = ["key", "n_cols", *group_flds]

    if not group_flds:
        raise ValueError("No group fields found")

    def _get_agg(col: str) -> str:
        mapper = {
            "start_ts": "min",
            "end_ts": "max",
            "round": "max",
            "duration_sec": "mean",
        }
        return mapper.get(
            col, "sum" if np.issubdtype(r_df.dtypes[col], np.number) else "last"
        )

    non_group_flds = [c for c in r_df.columns if c not in group_flds]
    agg_df = (
        r_df.groupby(group_flds)
        .aggregate(
            {c: _get_agg(c) for c in non_group_flds},
            total=("duration_sec", "count"),
        )
        .drop(columns=["pid", "host", "param_names", "round", "duration_sec"])
        .sort_index()
    )
    agg_df["duration_sec"] = (
        r_df.groupby(group_flds)
        .apply(calc_total_time_coverage_seconds, include_groups=False)
        .sort_index()
    )
    agg_df["M_rows_per_sec"] = (agg_df["n_rows"] / agg_df["duration_sec"] / 1e6).astype(
        "float64"
    )
    agg_df["MB_per_sec"] = (
        agg_df["n_bytes"] / 1024 ** 2 / agg_df["duration_sec"]
    ).astype("float64")

    return agg_df
